﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;
using System.Linq;

namespace RCM {
    public partial class MainForm : Form {
        #region Form
        private bool allowClose = false;

        protected override CreateParams CreateParams {
            // Workaround for control flickering reduction
            get {
                CreateParams cp = base.CreateParams;
                cp.ExStyle |= 0x02000000;  // WS_EX_COMPOSITED
                return cp;
            }
        }

        public MainForm() {
            InitializeComponent();
            treeView.ImageList = Config.IconList;
            treeView.ContextMenuStrip = GetNewRecordContextMenu();
            treeView.TreeViewNodeSorter = new RecordNodeSorter();
            LoadDataSources(0);
        }

        private void MainForm_Resize(object sender, EventArgs e) {
            if (WindowState == FormWindowState.Minimized) {
                ShowInTaskbar = false;
                return;
            }
            Config.Save();
        }

        private void NotifyIcon_MouseClick(object sender, MouseEventArgs e) {
            if (e.Button == MouseButtons.Left) {
                WindowState = FormWindowState.Normal;
                ShowInTaskbar = true;
            }
        }

        private void ToolStripMenuShowMainForm_Click(object sender, EventArgs e) {
            WindowState = FormWindowState.Normal;
            ShowInTaskbar = true;
        }

        private void ToolStripMenuCloseMainForm_Click(object sender, EventArgs e) {
            allowClose = true;
            Close();
        }

        private void LoadDataSources(int selectedIndex) {
            comboBoxDatabase.Items.Clear();
            comboBoxDatabase.Items.AddRange(Config.DataSources.ToArray());
            if (comboBoxDatabase.Items.Count == 0) {
                treeView.Enabled = false;
                treeView.Nodes.Clear();
            } else {
                treeView.Enabled = true;
                comboBoxDatabase.SelectedIndex = selectedIndex != -1 ? selectedIndex : 0;
            }
        }

        private void ButtonConfig_Click(object sender, EventArgs e) {
            int selectedIndex = comboBoxDatabase.SelectedIndex;
            ConfigForm configForm = new ConfigForm();
            configForm.ShowDialog();
            configForm.Dispose();
            LoadDataSources(selectedIndex);
        }

        private void MainForm_FormClosing(object sender, FormClosingEventArgs e) {
            if (e.CloseReason == CloseReason.UserClosing && !allowClose) {
                e.Cancel = true;
                WindowState = FormWindowState.Minimized;
            }
        }
        #endregion

        #region ComboBox & DataSource
        private IDataSource previouslySelectedDataSource = null;

        private void RebuildGroupCombobox(IRecord record) {
            List<GroupPath> groupPaths = new List<GroupPath>() { new GroupPath(null) };
            groupPaths.AddRange(RebuildGroupCombobox(treeView.Nodes));
            comboBoxGroup.Items.Clear();
            comboBoxGroup.Items.AddRange(groupPaths.ToArray());
            comboBoxGroup.SelectedItem = groupPaths.First(gp => gp.Id == record.GroupId);
        }

        private List<GroupPath> RebuildGroupCombobox(TreeNodeCollection nodes) {
            List<GroupPath> groupPaths = new List<GroupPath>();
            foreach (TreeNode node in nodes) {
                if ((IRecord)node.Tag is Group && !ReferenceEquals(node, treeView.SelectedNode)) {
                    groupPaths.Add(new GroupPath(node));
                    groupPaths.AddRange(RebuildGroupCombobox(node.Nodes));
                }
            }
            return groupPaths;
        }

        private void ShowDataSourceException(string failedAction, Exception ex) {
            IDataSource dataSource = (IDataSource)comboBoxDatabase.SelectedItem;
            string message = String.Format("{0}\n\nData source \"{1}\" exception:\n{2}\n{3}", failedAction, dataSource.Title, ex.GetType().ToString(), ex.Message);
            MessageBox.Show(message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
        }

        private void ComboBoxDatabase_SelectedIndexChanged(object sender, EventArgs e) {
            if (comboBoxDatabase.SelectedIndex == -1) {
                return;
            }
            try {
                treeView.Enabled = true;
                ReloadTreeView(true);
                previouslySelectedDataSource = (IDataSource)comboBoxDatabase.SelectedItem;
                treeView.SelectedNode = null;
                ChangeSelectedRecord();
            } catch (Exception ex) {
                ShowDataSourceException("Failed to load data source.", ex);
                if (previouslySelectedDataSource == null) {
                    treeView.Enabled = false;
                    treeView.Nodes.Clear();
                    comboBoxDatabase.SelectedIndex = -1;
                    toolStripStatusLabel.Text = "Last data source refresh: never";
                    toolStripStatusFailedFlag.Text = "(data source not loaded)";
                } else {
                    comboBoxDatabase.SelectedItem = previouslySelectedDataSource;
                }
            }
        }

        private void RefreshContextMenuItem_Click(object sender, EventArgs e) {
            try {
                ReloadTreeView(true);
            } catch (Exception ex) {
                ShowDataSourceException("Failed to refresh data source.", ex);
                toolStripStatusFailedFlag.Text = "(failed to refresh since then)";
            }
        }

        private void TimerDataSourceRefresh_Tick(object sender, EventArgs e) {
            try {
                ReloadTreeView();
            } catch (Exception ex) {
                if (toolStripStatusFailedFlag.Text == String.Empty) {
                    ShowDataSourceException("Failed to refresh data source.", ex);
                    toolStripStatusFailedFlag.Text = "(failed to refresh since then)";
                }
            }
        }
        #endregion

        #region TreeView display

        private void ReloadTreeView(bool restartTimer = false) {
            // Select all records from database
            IDataSource dataSource = (IDataSource)comboBoxDatabase.SelectedItem;
            Dictionary<Guid, IRecord> dataSourceRecords = dataSource.SelectAll();
            Guid selectedGuid = treeView.SelectedNode == null ? Guid.Empty : ((IRecord)treeView.SelectedNode.Tag).Id;

            // Flatten existing tree nodes
            List<TreeNode> flattenedTreeList = new List<TreeNode>();
            int lastProcessPosition = 0;
            flattenedTreeList.AddRange(treeView.Nodes.Cast<TreeNode>());
            while (lastProcessPosition < flattenedTreeList.Count) {
                flattenedTreeList.AddRange(flattenedTreeList[lastProcessPosition++].Nodes.Cast<TreeNode>());
            }
            Dictionary<Guid, TreeNode> flattenedTree = flattenedTreeList.ToDictionary(n => ((IRecord)n.Tag).Id, n => n);

            // Iterate over existing nodes
            HashSet<Guid> removeFromFlattenedTree = new HashSet<Guid>();
            foreach (TreeNode node in flattenedTree.Values) {
                IRecord record = (IRecord)node.Tag;
                // Delete node if record no longer present in database
                if (!dataSourceRecords.ContainsKey(record.Id)) {
                    node.Remove();
                    removeFromFlattenedTree.Add(record.Id);
                    continue;
                }
                // Update existing record if it has changed since the last refresh
                IRecord updatedRecord = dataSourceRecords[record.Id];
                if (ConfigSerializer.ObjToString(record) != ConfigSerializer.ObjToString(updatedRecord)) {
                    node.Tag = updatedRecord;
                    node.Text = updatedRecord.Title;
                    if (updatedRecord.GroupId != record.GroupId) {
                        node.Remove();
                        if (updatedRecord.GroupId == Guid.Empty) {
                            treeView.Nodes.Add(node);
                        } else {
                            flattenedTree[updatedRecord.GroupId].Nodes.Add(node);
                        }
                    }
                }
                // Remove processed record
                dataSourceRecords.Remove(record.Id);
            }
            // Remove deleted records from flattened tree
            foreach (Guid removedId in removeFromFlattenedTree) {
                flattenedTree.Remove(removedId);
            }

            // Create nodes from unprocessed (new) records
            foreach (IRecord record in dataSourceRecords.Values) {
                flattenedTree.Add(record.Id, CreateTreeNodeFromRecord(record));
            }

            // Assign nodes to tree structure (all nodes have to exist first)
            foreach (IRecord record in dataSourceRecords.Values) {
                TreeNode node = flattenedTree[record.Id];
                if (record.GroupId == Guid.Empty) {
                    treeView.Nodes.Add(node);
                } else {
                    flattenedTree[record.GroupId].Nodes.Add(node);
                }
            }
            // Expand nodes (all nodes have to be placed in tree first)
            foreach (IRecord record in dataSourceRecords.Values) {
                if (dataSource.ExpandedNodes.Contains(record.Id)) {
                    TreeNode node = flattenedTree[record.Id];
                    node.Expand();
                }
            }

            // Remove controls if selected node has been deleted
            if (!flattenedTree.ContainsKey(selectedGuid)) {
                SortTreeAndSelectNode(null);
            }

            // Restart timer
            toolStripStatusLabel.Text = String.Format("Last data source refresh: {0}", DateTime.Now.ToString("yyyy-MM-dd HH:mm"));
            toolStripStatusFailedFlag.Text = String.Empty;
            if (restartTimer) {
                timerDataSourceRefresh.Stop();
                timerDataSourceRefresh.Start();
            }

            // Delete expanded nodes which no longer exist
            HashSet<Guid> removeFromExpandedNodes = new HashSet<Guid>();
            foreach (Guid recordId in dataSource.ExpandedNodes) {
                if (!flattenedTree.ContainsKey(recordId)) {
                    removeFromExpandedNodes.Add(recordId);
                }
            }
            foreach (Guid recordId in removeFromExpandedNodes) {
                dataSource.ExpandedNodes.Remove(recordId);
            }

            // Save expanded nodes
            Config.Save();
        }

        private TreeNode CreateTreeNodeFromRecord(IRecord record) {
            string imageKey = record.GetType().FullName;
            return new TreeNode(record.Title) { Tag = record, ImageKey = imageKey, SelectedImageKey = imageKey, ContextMenuStrip = GetContextMenu(record) };
        }

        private void SortTreeAndSelectNode(TreeNode node) {
            /*if (node == null) {
                return;
            }*/
            treeView.Sort();
            if (treeView.SelectedNode == node) {
                ChangeSelectedRecord();
            } else {
                treeView.SelectedNode = node;
            }
        }
        #endregion

        #region TreeView drag & drop
        private void TreeView_ItemDrag(object sender, ItemDragEventArgs e) => DoDragDrop(e.Item, DragDropEffects.Move);

        private void TreeView_DragEnter(object sender, DragEventArgs e) => e.Effect = DragDropEffects.Move;

        private void TreeView_DragDrop(object sender, DragEventArgs e) {
            TreeNode draggedNode = (TreeNode)e.Data.GetData(typeof(TreeNode));
            if (draggedNode == null) {
                return;
            }
            TreeNode targetNode = treeView.GetNodeAt(treeView.PointToClient(new Point(e.X, e.Y)));
            bool canDrop = true;
            if (targetNode != null) {
                TreeNode parentNode = targetNode;
                canDrop = !draggedNode.Equals(targetNode) && targetNode != null && (IRecord)targetNode.Tag is Group;
                while (canDrop && (parentNode != null)) {
                    canDrop = !ReferenceEquals(draggedNode, parentNode);
                    parentNode = parentNode.Parent;
                }
            }
            if (canDrop) {
                IRecord record = (IRecord)draggedNode.Tag;
                record.GroupId = targetNode == null ? Guid.Empty : ((IRecord)targetNode.Tag).Id;
                try {
                    ((IDataSource)comboBoxDatabase.SelectedItem).Update(record);
                } catch {
                    MessageBox.Show("Failed to update record in data source", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }
                draggedNode.Remove();
                if (targetNode == null) {
                    treeView.Nodes.Add(draggedNode);
                } else {
                    targetNode.Nodes.Add(draggedNode);
                }
                SortTreeAndSelectNode(draggedNode);
            }
        }
        #endregion

        #region TreeView navigation & selection
        private void TreeView_AfterSelect(object sender, TreeViewEventArgs e) => ChangeSelectedRecord();

        private void TreeView_NodeMouseClick(object sender, TreeNodeMouseClickEventArgs e) => treeView.SelectedNode = e.Node;

        private void TreeView_MouseUp(object sender, MouseEventArgs e) {
            if (treeView.HitTest(e.X, e.Y).Node == null) {
                treeView.SelectedNode = null;
                ChangeSelectedRecord();
            }
        }

        private void TreeView_KeyDown(object sender, KeyEventArgs e) {
            if (e.KeyCode == Keys.Return) {
                if (treeView.SelectedNode != null) {
                    RunDefaultNodeAction(treeView.SelectedNode);
                }
            } else if (e.KeyCode == Keys.Delete) {
                if (treeView.SelectedNode != null) {
                    DeleteSelectedNode();
                }
            } else if (e.KeyCode == Keys.F2) {
                if (treeView.SelectedNode != null) {
                    treeView.SelectedNode.BeginEdit();
                }
            } else if (e.KeyCode == Keys.F5) {
                ReloadTreeView(true);
            }
        }

        private void RunDefaultNodeAction(TreeNode node) {
            IAction[] actions = ((IRecord)node.Tag).GetActions();
            if (actions.Length == 0) {
                return;
            }
            actions[0].Run();
        }

        private void TreeView_NodeMouseDoubleClick(object sender, TreeNodeMouseClickEventArgs e) {
            RunDefaultNodeAction(e.Node);
        }

        private void ChangeSelectedRecord() {
            bool isSelected = treeView.SelectedNode != null;
            textBoxTitle.Enabled = isSelected;
            textBoxTitle.Text = "";
            comboBoxGroup.Enabled = isSelected;
            comboBoxGroup.SelectedItem = null;
            buttonSave.Enabled = isSelected;
            LoadRecordControl();
        }

        private void LoadRecordControl() {
            foreach (Control control in panelRecord.Controls) {
                panelRecord.Controls.Remove(control);
                control.Dispose();
            }
            if (treeView.SelectedNode != null) {
                IRecord record = (IRecord)treeView.SelectedNode.Tag;
                // Prevent refresh of currently selected node (which is possibly being edited)
                //if (record.Id == panelRecord.
                textBoxTitle.Text = record.Title;
                RebuildGroupCombobox(record);
                panelRecord.Controls.Add(record.GetControl());
            }
        }

        private void TreeView_AfterExpand(object sender, TreeViewEventArgs e) {
            ((IDataSource)comboBoxDatabase.SelectedItem).ExpandedNodes.Add(((IRecord)e.Node.Tag).Id);
        }

        private void TreeView_AfterCollapse(object sender, TreeViewEventArgs e) {
            ((IDataSource)comboBoxDatabase.SelectedItem).ExpandedNodes.Remove(((IRecord)e.Node.Tag).Id);
        }
        #endregion

        #region Record create/update/delete
        public delegate void AfterLabelEditWorkaround(TreeNode node);

        private void TreeView_AfterLabelEdit(object sender, NodeLabelEditEventArgs e) {
            if (e.Label != null) {
                IRecord record = (IRecord)e.Node.Tag;
                record.Title = e.Label;
                try {
                    ((IDataSource)comboBoxDatabase.SelectedItem).Update(record);
                    // Workaround for treeView.Sort() triggering relabel on every sorted node. See note in https://docs.microsoft.com/en-us/dotnet/api/system.windows.forms.listview.afterlabeledit?view=netframework-4.6
                    BeginInvoke(new AfterLabelEditWorkaround(SortTreeAndSelectNode), new object[] { e.Node });
                } catch {
                    MessageBox.Show("Failed to update record in data source.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            } else {
                e.CancelEdit = true;
            }
        }

        private void ButtonSave_Click(object sender, EventArgs e) {
            TreeNode selectedNode = treeView.SelectedNode;
            IRecord record = (IRecord)selectedNode.Tag;
            record.UpdateFromControl();
            record.Title = selectedNode.Text = textBoxTitle.Text;
            GroupPath groupPath = (GroupPath)comboBoxGroup.SelectedItem;
            bool groupChanged = record.GroupId != groupPath.Id;
            record.GroupId = groupPath.Id;
            try {
                ((IDataSource)comboBoxDatabase.SelectedItem).Update(record);
            } catch {
                MessageBox.Show("Failed to update record in data source.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }
            if (groupChanged) {
                selectedNode.Remove();
                if (groupPath.Node != null) {
                    groupPath.Node.Nodes.Add(selectedNode);
                } else {
                    treeView.Nodes.Add(selectedNode);
                }
            }
            SortTreeAndSelectNode(selectedNode);
        }

        private void NewRecordContextMenuItem_Click(object sender, EventArgs e) {
            Type type = (Type)((ToolStripMenuItem)sender).Tag;
            TreeNode parentNode = treeView.SelectedNode;
            IRecord record = (IRecord)Activator.CreateInstance(type);
            record.Id = Guid.NewGuid();
            record.Title = string.Format("New {0}", type.GetDisplayName());
            record.GroupId = parentNode == null ? Guid.Empty : ((IRecord)parentNode.Tag).Id;
            try {
                TreeNode node = InsertNewRecord(record, parentNode);
                SortTreeAndSelectNode(node);
            } catch (Exception ex) {
                ShowDataSourceException("Failed to insert record into data source.", ex);
            }
        }

        private void DuplicateRecordContextMenuItem_Click(object sender, EventArgs e) {
            try {
                TreeNode node = DuplicateRecord(treeView.SelectedNode, treeView.SelectedNode.Parent);
                SortTreeAndSelectNode(node);
            } catch (Exception ex) {
                ShowDataSourceException("Failed to insert record into data source.", ex);
            }
        }

        private TreeNode DuplicateRecord(TreeNode sourceNode, TreeNode parentNode) {
            IRecord source = (IRecord)sourceNode.Tag;
            IRecord record = (IRecord)ConfigSerializer.JsonToObj(ConfigSerializer.ObjToJson(source));
            record.Id = Guid.NewGuid();
            record.Title = string.Format("{0} - Copy", record.Title);
            record.GroupId = parentNode == null ? Guid.Empty : ((IRecord)parentNode.Tag).Id;
            TreeNode node = InsertNewRecord(record, parentNode);
            if (record is Group) {
                foreach (TreeNode sourceChildNode in sourceNode.Nodes) {
                    DuplicateRecord(sourceChildNode, node);
                }
            }
            return node;
        }

        private TreeNode InsertNewRecord(IRecord record, TreeNode parentNode) {
            ((IDataSource)comboBoxDatabase.SelectedItem).Insert(record);
            TreeNode node = CreateTreeNodeFromRecord(record);
            if (parentNode != null) {
                parentNode.Nodes.Add(node);
            } else {
                treeView.Nodes.Add(node);
            }
            return node;
        }

        private void DeleteSelectedNode() {
            IRecord record = (IRecord)treeView.SelectedNode.Tag;
            string message = string.Format("Do you really want to delete selected item{0}?", record is Group ? " and all its subitems" : "");
            if (MessageBox.Show(message, "Confirmation", MessageBoxButtons.YesNo, MessageBoxIcon.Question) == DialogResult.Yes) {
                try {
                    ((IDataSource)comboBoxDatabase.SelectedItem).Delete(record);
                } catch (Exception ex) {
                    ShowDataSourceException("Failed to remove record from data source.", ex);
                    return;
                }
                treeView.SelectedNode.Remove();
                SortTreeAndSelectNode(null);
            }
        }

        private void DeleteRecordContextMenuItem_Click(object sender, EventArgs e) => DeleteSelectedNode();
        #endregion

        #region Record ContextMenu
        private ContextMenuStrip GetNewRecordContextMenu() {
            ContextMenuStrip menu = new ContextMenuStrip();
            ToolStripMenuItem menuItemNew = new ToolStripMenuItem("New item");
            // Add group as the first type
            Type groupType = typeof(Group);
            ToolStripMenuItem groupItem = new ToolStripMenuItem(groupType.GetDisplayName()) { Tag = groupType, Image = Config.IconList.Images[groupType.FullName] };
            groupItem.Click += NewRecordContextMenuItem_Click;
            menuItemNew.DropDownItems.Add(groupItem);
            menuItemNew.DropDownItems.Add(new ToolStripSeparator());
            // Add all the rest
            foreach (Type type in Config.RecordTypes) {
                if (type == groupType) {
                    continue;
                }
                ToolStripMenuItem item = new ToolStripMenuItem(type.GetDisplayName()) { Tag = type, Image = Config.IconList.Images[type.FullName] };
                item.Click += NewRecordContextMenuItem_Click;
                menuItemNew.DropDownItems.Add(item);
            }
            menu.Items.Add(menuItemNew);
            ToolStripMenuItem menuItemRefresh = new ToolStripMenuItem("Refresh data source");
            menuItemRefresh.Click += RefreshContextMenuItem_Click;
            menu.Items.Add(menuItemRefresh);
            return menu;
        }

        private ContextMenuStrip GetContextMenu(IRecord record) {
            ContextMenuStrip menu = record is Group ? GetNewRecordContextMenu() : new ContextMenuStrip();
            bool isDefaultAction = true;
            foreach (IAction action in record.GetActions()) {
                ToolStripMenuItem actionItem = new ToolStripMenuItem(action.GetType().GetDisplayName()) { Tag = action };
                if (isDefaultAction) {
                    actionItem.Font = actionItem.Font = new Font(actionItem.Font, actionItem.Font.Style | FontStyle.Bold);
                    isDefaultAction = false;
                }
                actionItem.Click += ActionItem_Click;
                menu.Items.Add(actionItem);
            }
            menu.Items.Add(new ToolStripSeparator());
            ToolStripMenuItem duplicateItem = new ToolStripMenuItem("Duplicate item");
            duplicateItem.Click += DuplicateRecordContextMenuItem_Click;
            menu.Items.Add(duplicateItem);
            ToolStripMenuItem deleteItem = new ToolStripMenuItem("Delete item");
            deleteItem.Click += DeleteRecordContextMenuItem_Click;
            menu.Items.Add(deleteItem);
            return menu;
        }

        private void ActionItem_Click(object sender, EventArgs e) {
            ((IAction)((ToolStripMenuItem)sender).Tag).Run();
        }
        #endregion

        /*private void button1_Click(object sender, EventArgs e) {
            List<IRecord> records = Importer.GetRecordsFromRDM();
            IDataSource dataSource = ((IDataSource)comboBoxDatabase.SelectedItem);
            foreach (IRecord record in records) {
                dataSource.Insert(record);
            }
        }*/
    }
}
