﻿using System;
using System.ComponentModel;
using System.Diagnostics;
using System.IO;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace RCM {
    [@DisplayName("Start Remote Desktop")]
    public class RDPAction : IAction {
        private RDP _rdp;
        private RDPConfig _config;

        public RDPAction(RDP rdp) {
            _rdp = rdp;
            _config = (RDPConfig)Config.ConfigSections[typeof(RDPConfig)];
        }

        public async void Run() {
            await Task.Run(() => {
                if (Environment.OSVersion.Platform == PlatformID.Unix) {
                    RunUnix();
                } else {
                    RunWin();
                }
            });
        }

        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
        private struct DATA_BLOB {
            public int cbData;
            public IntPtr pbData;
        }

        [DllImport("crypt32.dll", SetLastError = true, CharSet = CharSet.Auto)]
        private static extern bool CryptProtectData(ref DATA_BLOB pDataIn, string szDataDescr, IntPtr pOptionalEntropy, IntPtr pvReserved, IntPtr pPromptStruct, int dwFlags, ref DATA_BLOB pDataOut);

        public static string EncryptPassword(string plainText) {
            if (plainText == null) {
                plainText = String.Empty;
            }
            byte[] plainBytes = Encoding.Unicode.GetBytes(plainText);

            DATA_BLOB plainBlob = new DATA_BLOB();
            DATA_BLOB cipherBlob = new DATA_BLOB();

            try {
                plainBlob.pbData = Marshal.AllocHGlobal(plainBytes.Length);
                if (plainBlob.pbData == IntPtr.Zero) {
                    throw new Exception("Plaintext DATA_BLOB allocation exception");
                }
                plainBlob.cbData = plainBytes.Length;
                Marshal.Copy(plainBytes, 0, plainBlob.pbData, plainBytes.Length);

                bool result = CryptProtectData(ref plainBlob, String.Empty, IntPtr.Zero, IntPtr.Zero, IntPtr.Zero, 1, ref cipherBlob);
                if (!result) {
                    throw new Exception("CryptProtectData exception", new Win32Exception(Marshal.GetLastWin32Error()));
                }
                byte[] cipherBytes = new byte[cipherBlob.cbData];
                Marshal.Copy(cipherBlob.pbData, cipherBytes, 0, cipherBlob.cbData);
                return BitConverter.ToString(cipherBytes).Replace("-", "");
            } catch (Exception ex) {
                throw new Exception("DPAPI encryption exception", ex);
            } finally {
                if (plainBlob.pbData != IntPtr.Zero) {
                    Marshal.FreeHGlobal(plainBlob.pbData);
                }

                if (cipherBlob.pbData != IntPtr.Zero) {
                    Marshal.FreeHGlobal(cipherBlob.pbData);
                }
            }
        }

        private string CreateRdpFile() {
            string domain = null;
            string username;
            string[] splitname = _rdp.Username.Split("\\".ToCharArray());
            if (splitname.Length == 2) {
                domain = splitname[0];
                username = splitname[1];
            } else {
                username = _rdp.Username;
            }

            string rdpName = string.Format("{0}.{1}", _rdp.Title, Guid.NewGuid().ToString());
            foreach (var c in Path.GetInvalidFileNameChars()) {
                rdpName = rdpName.Replace(c, '_');
            }
            string rdpFile = Path.Combine(Path.GetTempPath(), string.Format("{0}.rdp", rdpName));
            // https://docs.microsoft.com/en-us/previous-versions/windows/it-pro/windows-server-2008-R2-and-2008/ff393699(v%3dws.10)
            using (StreamWriter sw = new StreamWriter(rdpFile)) {
                sw.Write("drivestoredirect:s:*\r\nredirectprinters:i:0\r\nredirectcomports:i:0\r\nredirectsmartcards:i:0\r\nredirectposdevices:i:0\r\nnetworkautodetect:i:0\r\nauthentication level:i:0\r\nprompt for credentials:i:0\r\nnegotiate security layer:i:1\r\npromptcredentialonce:i:0\r\nsmart sizing:i:0\r\nuse multimon:i:0\r\nbandwidthautodetect:i:1\r\n");
                sw.Write(string.Format("full address:s:{0}\r\n", _rdp.Host));
                sw.Write(string.Format("username:s:{0}\r\n", username));
                if (!string.IsNullOrEmpty(domain)) {
                    sw.Write(string.Format("domain:s:{0}\r\n", domain));
                }
                if (Environment.OSVersion.Platform == PlatformID.Unix) {
                    sw.Write(string.Format("password:s:{0}\r\n", _rdp.Password));
                } else {
                    sw.Write(string.Format("password 51:b:{0}\r\n", EncryptPassword(_rdp.Password)));
                }
            }
            return rdpFile;
        }

        private void RunWin() {
            string rdpFile = CreateRdpFile();
            Process.Start("mstsc", string.Format("\"{0}\"", rdpFile));
            Thread.Sleep(1000);
            File.Delete(rdpFile);
        }

        private void RunUnix() {
            string rdpFile = CreateRdpFile();
            string[] cmd = _config.LinuxRDPCommand.Replace("$RDP", rdpFile).Split(" ".ToCharArray(), 2);
            if (cmd.Length == 0) {
                MessageBox.Show("No RDP command was set", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            } else if (cmd.Length == 1) {
                Process.Start(cmd[0]);
            } else {
                Process.Start(cmd[0], cmd[1]);
            }
            Thread.Sleep(1000);
            File.Delete(rdpFile);
        }
    }
}
